#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import logging
import re

from manrododex.apiadapter import ApiAdapter
from manrododex.exceptions import NoneUUID, LangNotAvail
from manrododex.manga_helpers import Chapters

MANGA_ENDPOINT = "/manga"
CHAPTER_ENDPOINT = "/chapter"
# the uuid is actually a guid.
# I can validate it with a regex I stole from https://stackoverflow.com/a/42048037
GUID_REGEX = "[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"


def _get_default_title(info_title):
    logging.debug("Using default title.")
    return list(info_title.values())[0]


def _handle_title(title_settings, info_title, info_alttitles):
    title = title_settings[0]
    if not title:
        if title_settings[-1]:
            title = _get_default_title(info_title)
        else:
            logging.debug("Using alternative title with language '%s'", title_settings[1])
            for i in info_alttitles:
                try:
                    title = i[title_settings[1]]
                    break
                except KeyError:
                    continue
            if not title:
                logging.debug("No altTitle found for '%s' using default one.", title_settings[1])
                title = _get_default_title(info_title)
        logging.debug("Manga title set successfully.")
    else:
        logging.debug("Manga title already given not changing it.")
    logging.debug("Name to be used '%s'", title)
    return title


def _gen_helper(var_gen):
    gen = None
    if var_gen and re.search("(([0-9]+)/([0-9]+))|([0-9]+)", var_gen):
        if "/" in var_gen:
            logging.debug("Generating range.")
            to_gen = var_gen.split("/")
            if int(to_gen[0]) < int(to_gen[-1]):
                gen = [str(i) for i in range(int(to_gen[0]), int(to_gen[-1]) + 1)]
                logging.debug("Range successfully generated.")
            else:
                logging.debug("Failed to generate range, are you sure the range is in ascending order from left to "
                              "right ?")
                gen = None
        else:
            gen = var_gen
    return gen


def _gen_list(selected_chapters):
    """should return 1 tuple with 2 lists:
            The first should be the volume(s).
            The second the chapter(s).
            use 'v{num}v' to mark the number as volume.
                '/' to make a range.
                ',' to start a new rule.
            e.g: v7v99 would be volume 7 chapter 99.
                 v1/3v1 would be chapter 1 from vol 1, 2 and 3.
                 1,4,6 would download chapter 1, 4 and 6 regardless for the volume.
                 v6v would download volume 6 entirely.
                 """
    vol = list()
    chap = list()
    for entry in selected_chapters.strip().split(","):
        try:
            vol_match = re.search("v((([0-9]+)/([0-9]+))|([0-9]+))v", entry).group()
            logging.debug("Entry successfully matched.")
            vol_gen = vol_match.strip("v")
            chap_gen = entry.replace(vol_match, "")
        except AttributeError:
            logging.debug("Entry did not match, defaulting to all volumes.")
            vol_gen = None
            chap_gen = entry
        vol_a = _gen_helper(vol_gen)
        chap_a = _gen_helper(chap_gen)
        vol.extend(vol_a) if type(vol_a) is list else vol.append(vol_a)
        chap.extend(chap_a) if type(chap_a) is list else chap.append(chap_a)
    vol = sorted(list(set(vol)))
    chap = sorted(chap)
    logging.debug("vol/chap list successfully generated.")
    return vol, chap


def _check_ext_url(chap_id):
    """check whether the chapter have an external URL or not.
        True means yes it have an external URL.
        False for no."""
    logging.debug("Checking for external url.")
    info = ApiAdapter.make_request("get", f"{CHAPTER_ENDPOINT}/{chap_id}")["data"]
    if info["attributes"]["externalUrl"] == "null":
        logging.debug("external url found.")
        return True
    else:
        logging.debug("external url not found.")
        return False


def _get_vol_list(info):
    return list(reversed(info.keys()))  # list from last to first volume this is why I reversed it.


def _get_chap_list(info, volumes_list):
    chapters_complete = list()
    for volume in volumes_list:
        chapters = info[volume]["chapters"]
        # workaround for f6e7ce00-e09c-4ed2-b806-eb2fdc7a5f60
        if type(chapters) is dict:
            chapters_list = list(reversed(chapters.keys()))
            chapters_complete.extend(chapters_list)
        else:
            continue
    return chapters_complete


def _get_vol_chap_list(info):
    volumes_list = _get_vol_list(info)
    chapters_complete = _get_chap_list(info, volumes_list)
    return volumes_list, chapters_complete


class Manga:
    """The Manga uhh...yeah

    Parameters :
    -------------
    url_uuid:
        The url or the uuid of the manga that needs to be fetched.
    lang:
        the language code in which the user wants to download the
        manga.
    """

    def __init__(self, url_uuid, lang):
        try:
            self.uuid = re.search(GUID_REGEX, url_uuid).group()
            logging.debug("the uuid extracted is : '%s'", self.uuid)
        except AttributeError:
            logging.critical("Failed to extract uuid skipping.")
            self.uuid = None
            raise NoneUUID("Failed to get uuid, execution cannot proceed.")
        else:
            self.info = dict()
            self.lang = lang
            self.chapters = Chapters()
            logging.info("Manga created with no errors.")

    def get_info(self, title_settings):
        """Gets more info about the manga, like:
        title, altTitles, description, status, availableTranslatedLanguages, year, contentRating

        Parameters:
        ------------
        title_settings :
        A tuple that contains 3 elements:
            The first:
                Type: str
                Description: a custom title for the manga.
                Default: None, which means to get it from the site.
            The second:
                Type: str
                Description: the language for the altTitle to be used.
                Default: None, which means to use the language of the manga.
            The third:
                Type: bool
                Description: force the usage of the default title - the one you see on the website - regardless of the
                language specified.
                Default: True

        Expected response a dictionary with the following keys that interests us:
        title: dict
        altTitles: list containing a dict
        description: dict
        status: str
        availableTranslatedLanguages: list
        year: int
        contentRating: str
        """

        info = ApiAdapter.make_request("get", f"{MANGA_ENDPOINT}/{self.uuid}")["data"]["attributes"]
        logging.debug("Checking if requested language is available.")
        avl_langs = info["availableTranslatedLanguages"]
        if self.lang not in avl_langs:
            logging.critical("Language not available skipping.")
            raise LangNotAvail("Requested language not available for this manga.")
        del avl_langs
        self.info["name"] = _handle_title(title_settings, info["title"], info["altTitles"])
        del title_settings
        try:
            self.info["description_formatted"] = info["description"][self.lang]
            logging.debug("Using description with language '%s'", self.lang)
        except KeyError:
            logging.debug("No description with language '%s'\n Using the first one in the request.", self.lang)
            try:
                self.info["description_formatted"] = next(iter(info["description"].values()))
                logging.debug("Using description with language '%s'", next(iter(info["description"])))
            except StopIteration:
                logging.critical("No suitable description found setting it to None.")
        self.info["status"] = info["status"]
        try:
            self.info["year"] = int(info["year"])
        except TypeError:
            self.info["year"] = info["year"]
        self.info["age_rating"] = info["contentRating"]
        logging.info("All info fetched successfully.")

    def _chapters_to_queue(self, vol_list, chap_list, info):
        for vol in vol_list:
            try:
                vol_tmp = info[vol]
                for chap in chap_list:
                    try:
                        chap_tmp = vol_tmp["chapters"][chap]
                        chap_id = None
                        if chap_tmp["count"] != 1 and _check_ext_url(chap_tmp["id"]):
                            for id_o in chap_tmp["others"]:
                                if not _check_ext_url(id_o):
                                    chap_id = id_o
                                    break
                        else:
                            chap_id = chap_tmp["id"]
                        if chap_id:
                            self.chapters.put((vol, chap, chap_id))
                            logging.debug("Item successfully added to queue.")
                        else:
                            logging.debug("no chapter id found skipping.")
                            continue
                    except KeyError:
                        continue
            except KeyError:
                continue

    def get_chapters(self, selected_vol_chap):
        """Make the request to get chapters by language and add them to the chapters SimpleQueue in the following
        format: (vol, chap, chap_id)"""
        info = ApiAdapter.make_request("get",
                                       f"{MANGA_ENDPOINT}/{self.uuid}/aggregate",
                                       passed_params={
                                           "translatedLanguage[]": f"{self.lang}"
                                       })["volumes"]
        # not sure if it's supposed to be fetch or download.
        if selected_vol_chap:
            logging.debug("Using user provided vol/chapters.")
            vol_list, chap_list = _gen_list(selected_vol_chap)
            del selected_vol_chap
            if len(vol_list) == len(chap_list) == 1 and vol_list[0] is chap_list[0] is None:
                # if both are None Download everything.
                logging.debug("Something with the generated vol/chap using all vol/chap.")
                vol_list, chap_list = _get_vol_chap_list(info)
            elif len(vol_list) == 1 and not vol_list[0]:
                # if volume is only None download all volumes and respect user given chapters.
                logging.debug("Using all vol and user given chap.")
                vol_list = _get_vol_list(info)
            elif len(chap_list) == 1 and not chap_list[0]:
                # if chapters is only None download all chapters and respect user given volumes.
                logging.debug("Using user given vol and downloading all chapters.")
                chap_list = _get_chap_list(info, vol_list)
            self._chapters_to_queue(vol_list, chap_list, info)
        else:
            logging.debug("Using all volumes and chapters.")
            vol_list, chap_list = _get_vol_chap_list(info)
            self._chapters_to_queue(vol_list, chap_list, info)
        logging.info("All chapters have been fetched and ready for the next step.")



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from tqdm import tqdm

from manrododex.downloader import Downloader
from manrododex.exceptions import NoneUUID, LangNotAvail, RequestDidNotSucceed
from manrododex.manga import Manga
from manrododex.system_helper import SysHelper


def main(url_uuid, title_settings, lang, selected_vol_chap, main_path, quality, threads, force_ssl, archive_format,
         dry_run):
    """A comment to keep track of the parameters.
    Parameters:
    ------------
    url_uuid:
        type: str
        default: No Default.
    title_settings:
        type: tuple
        default: (None, None, True)
    lang:
        type: str
        default: 'en'
    selected_vol_chap:
        type: str
        default: None
    main_path:
        type: str/path
        default: current working dir + 'Manga'
    quality:
        type: str
        default: data
        other option(s): data-saver
    threads:
        type: int
        default: 1
    force_ssl:
        type: bool
        default: False
        other option(s): True
    archive_format:
        type: str
        default: cbz
        other option(s): zip
    log_level:
        type: str
        default: INFO
    dry_run:
        type: bool
        default: False
    """
    try:
        main_bar = tqdm(desc="Starting Manga Download",
                        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} chapters",
                        position=0,
                        leave=False)
        manga = Manga(url_uuid, lang)
        main_bar.desc = f"Getting info on {manga.uuid} with language {manga.lang}"
        main_bar.refresh()
        del url_uuid, lang
        # Make the request to get basic info about the manga.
        manga.get_info(title_settings)
        main_bar.desc = f"Downloading : {manga.info['name']}"
        main_bar.refresh()
        del title_settings
        # Make the requests to get the available chapters.
        manga.get_chapters(selected_vol_chap)
        main_bar.total = manga.chapters.qsize()
        main_bar.desc = f"Downloading {manga.info['name']} chapters"
        main_bar.refresh()
        del selected_vol_chap
        # It's now time to download the manga.
        sys_helper = SysHelper(main_path, manga.info["name"], archive_format, dry_run)
        del main_path, archive_format
        # First create the main manga directory where the manga need to be put.
        sys_helper.create_main_manga_dir()
        # Then we create the directory for the manga that have the manga title as name.
        sys_helper.create_manga_dir()
        # Create the series.json
        sys_helper.write_series_json(manga.info)
        del manga.info
        # We can now start downloading.
        downloader = Downloader(manga.chapters, quality, threads, force_ssl, dry_run)
        del manga, quality, threads, force_ssl
        downloader.main(sys_helper, main_bar)
    except (NoneUUID, LangNotAvail, RequestDidNotSucceed):
        return 1

    return 0


def file_main(file, title_settings, lang, selected_vol_chap, main_path, quality, threads, force_ssl, archive_format,
              dry_run):
    """The file should contain at most three things the first being mandatory:
     url/uuid
     custom manga name
     the language"""
    with open(file, "r") as f:
        lines = f.readlines()
        for line in lines:
            line_comp = list(map(lambda x: str(x.strip().strip(" ")), line.strip().split(",")))
            uuid_url = line_comp[0]
            manga_name = line_comp[1]
            language = line_comp[2] if len(line_comp[2]) == 2 or len(line_comp[2]) == 5 else lang
            if len(line_comp) == 3:
                main(uuid_url, (manga_name, None, False), language, selected_vol_chap, main_path, quality,
                     threads, force_ssl, archive_format, dry_run)
            elif len(line_comp) == 2:
                main(uuid_url, (manga_name, None, False), lang, selected_vol_chap, main_path, quality, threads,
                     force_ssl, archive_format, dry_run)
            elif len(line_comp) == 1:
                main(uuid_url, title_settings, lang, selected_vol_chap, main_path, quality, threads, force_ssl,
                     archive_format, dry_run)



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import json
import logging
from os import path, makedirs, listdir, remove
from shutil import rmtree
from zipfile import ZipFile


def path_exits(given_path):
    return path.exists(given_path)


class SysHelper:
    def __init__(self, main_path, manga_title, archive_format, dry_run):
        self.main_path = main_path
        self.manga_path = path.join(self.main_path, manga_title)
        self.chapter_path = None
        self.archive_format = archive_format
        self.archive_path = None
        self.dry_run = dry_run
        logging.info("SysHelper created with no errors.")

    def create_main_manga_dir(self):
        if self.dry_run:
            logging.debug("Dry run not creating directory.")
            return
        logging.debug("Creating main manga directory at %s", self.main_path)
        if not path.exists(self.main_path):
            makedirs(self.main_path)
        logging.debug("Directory already exists, not creating.")

    def create_manga_dir(self):
        if self.dry_run:
            logging.debug("Dry run not creating directory.")
            return
        logging.debug("Creating main manga directory at %s", self.manga_path)
        if not path.exists(self.manga_path):
            makedirs(self.manga_path)
        logging.debug("Directory already exists, not creating.")

    def create_chapter_dir(self, chapter_name):
        self.chapter_path = path.join(self.manga_path, chapter_name)
        if self.dry_run:
            logging.debug("Dry run not creating directory.")
            return
        logging.debug("Creating main manga directory at %s", self.chapter_path)
        if not path.exists(self.chapter_path):
            makedirs(self.chapter_path)
        logging.debug("Directory already exists, not creating.")

    def forge_img_path(self, img_name, img_ext):
        return path.join(self.chapter_path, f"{img_name}{img_ext}")

    # noinspection PyTypeChecker
    def archive_chapter(self, any_to_fix):
        if self.dry_run:
            logging.debug("Dry run not archiving chapter.")
            return
        if not any_to_fix:
            logging.debug("Archiving directory.")
            archive_mode = "w"
        else:
            # append images to zip.
            logging.debug("Appending images to already existing archive.")
            archive_mode = "a"
        with ZipFile(self.archive_path, mode=archive_mode) as f:
            dir_list = self.get_dir_content()
            for file in dir_list:
                file_path = path.join(self.chapter_path, file)
                logging.debug("Adding file %s", file_path)
                f.write(file_path, path.basename(file_path))
        logging.info("All images added to archive successfully.")
        self.del_dir()

    def check_if_already_exists(self, chapter_name):
        self.chapter_path = path.join(self.manga_path, chapter_name)
        self.archive_path = self.chapter_path + f".{self.archive_format}"
        if self.dry_run:
            logging.debug("Dry run not checking if it already exits.")
            return
        if path.exists(self.chapter_path) and path.exists(self.archive_path):
            logging.info("Folder and Archive already exists !!")
            return "FolderAndArchive"
        elif path.exists(self.chapter_path):
            logging.info("Folder already exists !!")
            return "Folder"
        elif path.exists(self.archive_path):
            logging.info("Archive already exists !!")
            return "Archive"
        else:
            return None

    def get_dir_content(self):
        if self.dry_run:
            logging.debug("Dry run not getting content.")
            return
        return listdir(self.chapter_path)

    def get_archive_content(self):
        if self.dry_run:
            logging.debug("Dry run not getting archive content.")
            return
        with ZipFile(self.archive_path, "r") as f:
            zip_content = f.namelist()
            return zip_content

    def del_archive(self):
        if self.dry_run:
            logging.debug("Dry run not deleting archive.")
            return
        logging.debug("Deleting Archive : %s", self.archive_path)
        remove(self.archive_path)

    def del_dir(self):
        if self.dry_run:
            logging.debug("Dry run not deleting directory.")
            return
        logging.debug("Deleting directory and all it's content : %s", self.chapter_path)
        rmtree(self.chapter_path)

    def write_series_json(self, info):
        if self.dry_run:
            logging.debug("Dry run not writing series.json.")
            return
        series_json = path.join(self.manga_path, "series.json")
        if path.exists(series_json):
            logging.debug("series.json already exists skipping.")
            return
        # format from https://github.com/mylar3/mylar3/wiki/series.json-examples
        data = {
            "metadata": info
        }
        with open(series_json, "w") as s:
            s.write(json.dumps(data, indent=4))
        logging.info("series.json has been written.")



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

from manrododex.exceptions import NoneUUID

uuid_urls = ("https://mangadex.org/title/66d82067-2117-4124-b54b-89b19c8bde45/the-idolm-ster-cinderella-girls-mizu-no"
             "-naka-no-tsubomi-doujinshi", "b98c4daf-be1a-46c8-ad83-21d532995240", "dasd")
ex_uuid_urls = ("66d82067-2117-4124-b54b-89b19c8bde45", "b98c4daf-be1a-46c8-ad83-21d532995240", NoneUUID)

ex_resp_main = (0, 0, 1)

gen = ("v7v99", "v1/3v1", "1,4,6", "v6v", "ffg", "v6/v")
ex_gen = ((["7"], ["99"]), (["1", "2", "3"], ["1"]), ([None], ["1", "4", "6"]), (["6"], [None]), ([None], [None]))

img_link = (
    "https://uploads.mangadex.org/data/3303dd03ac8d27452cce3f2a882e94b2/1-f7a76de10d346de7ba01786762ebbedc666b412ad0d4b73baa330a2a392dbcdd.png",
    "https://uploads.mangadex.org/data-saver/3303dd03ac8d27452cce3f2a882e94b2/1-27e7476475e60ad4cc4cefdb9b2dce29d84f490e145211f6b2e14b13bdb57f33.jpg",
    "https://uploads.mangadex.org/data-saver/e54e950ab0b830ff3036d2be72574c1e/26-36abab0bf677ff1b977e3a14dbdda37acf69ebcb0b4efe3979574edafe1a455d.jpg",
    "x1-b765e86d5ecbc932cf3f517a8604f6ac6d8a7f379b0277a117dc7c09c53d041e.png",
    "x18-41d36bffc607b627590bed2ce328b5dd87301bfc54fac25a96bec19160039c2e.jpg")
ex_img_name = ("1", "1", "26", "1", "18")
ex_img_exte = (".png", ".jpg", ".jpg", ".png", ".jpg")

g_lang = ("en", "aa", "ar", "ja", "pt-br", "12", "p")
ex_lang = ("en", "aa", "ar", "ja", "pt-br", "en", "en")



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import re

import pytest

import test

img_l = test.img_link
ex_img_exte = test.ex_img_exte
param = [(u, v) for u, v in zip(img_l, ex_img_exte)]


@pytest.mark.parametrize("img_link,ex_img_ext", param)
def test_download_image(img_link, ex_img_ext):
    img_ext = re.search("(-)(.*)(\..*$)", img_link).group(3)
    assert img_ext == ex_img_ext



#  Copyright (c) 2022 Charbel Assaad
# 
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
# 
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
# 
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import os
from tempfile import gettempdir

# https://github.com/qtile/qtile/blob/master/libqtile/scripts/main.py
try:
    # Python>3.7 can get the version from importlib
    from importlib.metadata import distribution

    VERSION = distribution("manrododex").version
except ModuleNotFoundError:
    try:
        # pkg_resources is required for 3.7
        import pkg_resources

        VERSION = pkg_resources.require("manrododex")[0].version
    except (pkg_resources.DistributionNotFound, ModuleNotFoundError):
        VERSION = "dev"


def check_logger(log):
    if log.upper() in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        return log.upper()
    else:
        return "INFO"


# https://github.com/manga-py/manga-py/blob/d89501a0f78d498f85114320d6123f59d328a905/manga_py/cli/_args_general.py#L4
def _gen_args(args_parser):
    args = args_parser.add_argument_group("General options")
    args.add_argument(
        "--version",
        action="version",
        version=VERSION,
        help=(
            "Shows %(prog)s's version and exit.\n"
        )
    )
    args.add_argument(
        "--log-level",
        type=check_logger,
        metavar="LOGLVL",
        default="INFO",
        help="Set the log level.\n"
             f"Default log file location {os.path.join(gettempdir(), 'manrododex.log')}\n"
             "possible values : 'DEBUG' 'INFO' 'WARNING' 'ERROR' 'CRITICAL'\n",
    )
    args.add_argument(
        "--lang-codes",
        action="store_true",
        default=False,
        help=(
            "Prints the available language codes and exit.\n"
        )
    )
    args.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help=(
            "Run the program without creating/saving anything.\n"
        )
    )



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import logging
import re
from time import sleep, time

import requests
from requests import JSONDecodeError
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from manrododex.exceptions import ResultNotOk, RequestDidNotSucceed

API_URL = "https://api.mangadex.org"
REPORT_ENDPOINT_URL = "https://api.mangadex.network/report"


class ApiAdapter:
    """Class to make the requests better ?
    well I'm lying here, I just want to objectify everything.
    """
    session = requests.session()
    retries = Retry(total=5,
                    backoff_factor=0.25,
                    status_forcelist=[500, 502, 503, 504])
    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.mount('https://', HTTPAdapter(max_retries=retries))
    lastu_taimo = None

    @classmethod
    def can_i_mauwku_requesto_senpai(cls):
        cuwuent_taimo = time()
        if cls.lastu_taimo:
            deuwuta = cuwuent_taimo - cls.lastu_taimo
            logging.debug("Time since last request '%s'", str(deuwuta))
            if deuwuta >= 0.2:
                cls.lastu_taimo = cuwuent_taimo
                return
            else:
                suweep_fuwu = 0.2 - deuwuta
                logging.debug("Waiting '%s'", str(suweep_fuwu))
                sleep(suweep_fuwu)
                cls.lastu_taimo = cuwuent_taimo
                return
        else:
            cls.lastu_taimo = cuwuent_taimo
            return

    # future me : Class methods are different -- they are called by a class, which is passed to the cls parameter of
    # the method. (Sololearn)
    @classmethod
    def make_request(cls, method, endpoint, passed_params=None, passed_headers=None):
        cls.can_i_mauwku_requesto_senpai()
        req = cls.session.request(method, f"{API_URL}{endpoint}", params=passed_params, headers=passed_headers)
        if req.status_code == 200:
            logging.debug("Request successful.")
            try:
                if req.json().get("result") == "ok":
                    return req.json()
                else:
                    raise ResultNotOk("Received response is invalid.")
            except JSONDecodeError:
                logging.error("Failed to decode json.")
                return None
        else:
            logging.error("Failed to make request.")
            raise RequestDidNotSucceed

    @classmethod
    def img_download(cls, img_link, img_path, bar, dry_run):
        should_report = True if not re.search(".*(\.?mangadex.org).*", img_link) else False
        cls.can_i_mauwku_requesto_senpai()
        logging.debug("Downloading img from link : %s", img_link)
        # https://stackoverflow.com/a/62113293
        if dry_run:
            logging.debug("Dry run not downloading img.")
            return
        req = cls.session.request("get", img_link, stream=True)
        success = req.status_code == 200
        if success:
            logging.debug("Request successful.")
            total = int(req.headers.get('content-length', 0))
            bar.total = total
            with open(img_path, "wb") as f:
                for data in req.iter_content(chunk_size=1024):
                    size = f.write(data)
                    bar.update(size)
        if should_report:
            try:
                cached = True if "HIT" in req.headers["X-Cache"] else False
            except KeyError:
                cached = False
            logging.debug("Reporting request to mangadex.")
            # The time function returns the time in seconds.
            # Mangadex wants it in milliseconds.
            duration = req.elapsed.microseconds
            data = {
                "url": img_link,
                "success": success,
                "cached": cached,
                "bytes": len(req.content),
                "duration": duration
            }
            header = {
                "content-type": "application/json"
            }
            # report
            cls.session.request("post", REPORT_ENDPOINT_URL, params=data, headers=header)
        if not success:
            logging.error("Failed to make request.")
            raise RequestDidNotSucceed



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import json
import sys

import manrododex.logger as logger
from manrododex.CLI.argparse import initialize_args
from manrododex.main import main, file_main


def print_lang_codes():
    from manrododex.mangadex_lang_codes import lang_codes
    print(json.dumps(lang_codes, indent=4))


def cli_handler():
    parser = initialize_args()
    args = parser.parse_args()
    if args.lang_codes:
        print_lang_codes()
        return
    if not args.dry_run:
        logger.init(args.log_level)
    title_settings = (args.name, args.alttitle_lang, args.deftitle)
    if len(sys.argv) <= 1:
        parser.print_help()
    else:
        if args.File is None:
            main(args.url_uuid, title_settings, args.language, args.selvolchap, args.destination, args.data_saver,
                 args.threads,
                 args.force_ssl, args.zip_format, args.dry_run)
        else:
            file_main(args.File, title_settings, args.language, args.selvolchap, args.destination, args.data_saver,
                      args.threads,
                      args.force_ssl, args.zip_format, args.dry_run)


if __name__ == "__main__":
    cli_handler()



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import logging
import re
from threading import Thread

from tqdm import tqdm

from manrododex.apiadapter import ApiAdapter
from manrododex.manga_helpers import Images

AT_HOME_SERVER_ENDPOINT = "/at-home/server"


def _chapter_archive_name(vol, chap):
    if vol != "none":
        chapter_name = f"vol-{vol}-chapter-{chap}"
    elif chap == "Oneshot":
        chapter_name = chap
    else:
        chapter_name = f"chapter-{chap}"
    return chapter_name


def _get_image_name(img_link):
    return re.search("(x?)([0-9]+)(-)", img_link).group(2)


def _get_image_ext(img_link):
    return re.search("(-)(.*)(\..*$)", img_link).group(3)


class Downloader:
    """This class is responsible for downloading the manga.
    Parameters :
    -------------
    manga:
        The Manga object to be downloaded.
    quality:
        The quality of the images to be used, data or data-saver.
    threads:
        The number of threads to be used.
    force_ssl:
        Force selecting from MangaDex@Home servers that use the standard HTTPS port 443.
        from https://api.mangadex.org/swagger.html
    """

    def __init__(self, chapters, quality, threads, force_ssl, dry_run):
        self.chapters = chapters
        self.quality = quality
        self.threads = threads
        self.force_ssl = force_ssl
        self.dry_run = dry_run
        self.images = Images()
        self.chapter_name = None
        logging.info("Downloader created successfully.")

    def _build_images_link(self, sys_helper):
        chapter = self.chapters.get()
        self.chapter_name = _chapter_archive_name(chapter[0], chapter[1])
        logging.debug("Chapter will be named : %s", self.chapter_name)
        exists = sys_helper.check_if_already_exists(self.chapter_name)
        img_check = None
        if exists == "FolderAndArchive":
            # Check Both and complete the one with the most images.
            dir_list = sys_helper.get_dir_content()
            archive_list = sys_helper.get_archive_content()
            if len(dir_list) >= len(archive_list):
                # if directory contains more images delete the archive.
                # or if they're equal delete the archive the directory is easier to handle.
                logging.info("Chapter Directory contains more stuff than the archive, deleting the archive.")
                del archive_list
                sys_helper.del_archive()
                img_check = dir_list
                del dir_list
                exists = None
            elif len(archive_list) > len(dir_list):
                # if archive contains more delete the directory
                logging.info("Archive contains more stuff than the directory, deleting the directory.")
                del dir_list
                sys_helper.del_dir()
                img_check = archive_list
                del archive_list
                exists = "Archive"
        elif exists == "Folder":
            # Folder already exists download the missing images if any.
            logging.info("Completing what's missing from the directory if any.")
            img_check = sys_helper.get_dir_content()
            exists = None
        elif exists == "Archive":
            # Archive already exists download the missing images if any.
            logging.info("Completing what's missing from the archive if any.")
            img_check = sys_helper.get_archive_content()
        sys_helper.create_chapter_dir(self.chapter_name)
        info = ApiAdapter.make_request("get",
                                       f"{AT_HOME_SERVER_ENDPOINT}/{chapter[2]}",
                                       passed_params={
                                           "forcePort443": self.force_ssl
                                       })
        del chapter
        base_url = info["baseUrl"]
        chapter_hash = info["chapter"]["hash"]
        images = info["chapter"]["dataSaver"] if self.quality == "data-saver" else info["chapter"]["data"]
        del info
        for image in images:
            if img_check:
                if f"{_get_image_name(image)}{_get_image_ext(image)}" in img_check:
                    continue
            self.images.put(f"{base_url}/{self.quality}/{chapter_hash}/{image}")
            logging.debug("Image added successfully.")
        logging.info("All required images for the chapter have been added.")
        return exists

    def _download_image(self, sys_helper, bar, thread_desc):
        img_link = self.images.get()
        img_name = _get_image_name(img_link)
        img_ext = _get_image_ext(img_link)
        bar.set_description(f"{thread_desc} Downloading {img_name}{img_ext}")
        img_path = sys_helper.forge_img_path(img_name, img_ext)
        del img_name, img_ext
        ApiAdapter.img_download(img_link, img_path, bar, self.dry_run)
        del img_link
        self.images.task_done()

    def _download_images(self, sys_helper, thread_number, chapter_bar):
        thread_desc = f"Thread nb {thread_number}"
        bar = tqdm(desc=thread_desc,
                   position=thread_number + 1,
                   unit="iB",
                   unit_scale=True,
                   unit_divisor=1024,
                   leave=False)
        while not self.images.empty():
            self._download_image(sys_helper, bar, thread_desc)
            chapter_bar.update(1)
            bar.reset()

    def main(self, sys_helper, main_bar):
        chapter_bar = tqdm(desc="Downloading chapters.",
                           bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} images",
                           position=1,
                           leave=False)
        while not self.chapters.empty():
            any_to_fix = self._build_images_link(sys_helper)
            total = self.images.qsize()
            chapter_bar.__dict__.update({"total": total,
                                         "desc": f"Downloading {self.chapter_name}"})
            if not self.images.empty():
                for i in range(self.threads):
                    Thread(target=self._download_images, args=(sys_helper, i + 1, chapter_bar)).start()
                self.images.join()
            sys_helper.archive_chapter(any_to_fix)
            chapter_bar.reset()
            main_bar.update(1)



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

# https://github.com/manga-py/manga-py/blob/d89501a0f78d498f85114320d6123f59d328a905/manga_py/cli/_args_downloading.py#L2
import os
import pathlib

from manrododex.mangadex_lang_codes import lang_codes


def check_lang(lang):
    if lang in lang_codes:
        return lang
    else:
        return "en"


def _download_args(args_parser):
    args = args_parser.add_argument_group("Downloading options")
    args.add_argument(
        "-d",
        "--destination",
        metavar="PATH",
        type=str,
        default=str(os.path.join(pathlib.Path().resolve().absolute(), "Manga")),
        help=(
            "Destination folder: where the manga will be saved.\n"
            "The path will be './%(metavar)s/manga_name/'\n"
        )
    )
    args.add_argument(
        "-t",
        "--threads",
        type=int,
        metavar="THREADS",
        default=1,
        help=(
            "Sets the number of threads to be used.\n"
            "Aka the number of images to be downloaded in parallel.\n"
            "Please don't use a lot of threads as it might get you banned.\n"
        ),
    )
    args.add_argument(
        "-l",
        "--language",
        type=check_lang,
        metavar="LANG",
        default="en",
        help="Set the language in which the chapters should be downloaded with.\n",
    )
    args.add_argument(
        "-ds",
        "--data-saver",
        action="store_const",
        const="data-saver",
        default="data",
        help=(
            "Switch the quality mode from 'data' to 'data-saver'.\n"
            "'data-saver' will download a compressed image instead of upload quality.\n"
        ),
    )
    args.add_argument(
        "-svc",
        "--selvolchap",
        type=str,
        metavar="SELVOLCHAP",
        default=None,
        help=(
            "Select chapters to be downloaded can be singles separated by ','\n"
            "use 'v{num}v' to mark the number as volume.\n"
            "    '/' to make a range.\n"
            "    ',' to start a new rule.\n"
            "e.g: v7v99 would be volume 7 chapter 99.\n"
            "     v1/3v1 would be chapter 1 from vol 1, 2 and 3.\n"
            "     1,4,6 would download chapter 1, 4 and 6 regardless for the volume.\n"
            "     v6v would download volume 6 entirely.\n"
        ),
    )
    args.add_argument(
        "--alttitle-lang",
        type=check_lang,
        metavar="ALTTLANG",
        default=None,
        help="Specify the language in we should get the alternative title.\n"
             "Available alternative titles can be seen to the left of the chapters of the manga on the site.\n",
    )
    args.add_argument(
        "--deftitle",
        action="store_false",
        default=True,
        help="Whether or not to use the default title of the manga.\n"
             "Aka the one that appears on the site/url.\n"
             "Note: this argument precedes the one of the alternative title in importance.\n"
             "By which i mean that if this argument is not used the program will get the default title even if"
             "the --alttitle-lang is used.\n",
    )
    args.add_argument(
        "--name",
        type=str,
        metavar="NAME",
        default=None,
        help="Set a custom name for the manga.\n"
             "Note: this argument precedes the both of the --alttitle-lang and --deftitle in importance\n",
    )
    args.add_argument(
        "--force-ssl",
        action="store_true",
        default=False,
        help="Force selecting from MangaDex@Home servers that use the standard HTTPS port 443.\n"
             "from https://api.mangadex.org/swagger.html\n",
    )
    args.add_argument(
        "--zip-format",
        action="store_const",
        const="zip",
        default="cbz",
        help="The use of this argument will switch it to zip.\n"
             "By default the cbz archive format will be used.\n",
    )



#  Copyright (c) 2022 Charbel Assaad
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

lang_codes = {
  "aa": "Afar",
  "ab": "Abkhazian",
  "ae": "Avestan",
  "af": "Afrikaans",
  "ak": "Akan",
  "am": "Amharic",
  "an": "Aragonese",
  "ar": "Arabic",
  "as": "Assamese",
  "av": "Avaric",
  "ay": "Aymara",
  "az": "Azerbaijani",
  "ba": "Bashkir",
  "be": "Belarusian",
  "bg": "Bulgarian",
  "bh": "Bihari languages",
  "bi": "Bislama",
  "bm": "Bambara",
  "bn": "Bengali",
  "bo": "Tibetan",
  "br": "Breton",
  "bs": "Bosnian",
  "ca": "Catalan |  Valencian",
  "ce": "Chechen",
  "ch": "Chamorro",
  "co": "Corsican",
  "cr": "Cree",
  "cs": "Czech",
  "cu": "Church Slavic |  Old Slavonic |  Church Slavonic |  Old Bulgarian |  Old Church Slavonic",
  "cv": "Chuvash",
  "cy": "Welsh",
  "da": "Danish",
  "de": "German",
  "dv": "Divehi |  Dhivehi |  Maldivian",
  "dz": "Dzongkha",
  "ee": "Ewe",
  "el": "Greek, Modern (1453-)",
  "en": "English",
  "eo": "Esperanto",
  "es": "Castilian Spanish",
  "et": "Estonian",
  "eu": "Basque",
  "fa": "Persian",
  "ff": "Fulah",
  "fi": "Finnish",
  "fj": "Fijian",
  "fo": "Faroese",
  "fr": "French",
  "fy": "Western Frisian",
  "ga": "Irish",
  "gd": "Gaelic |  Scottish Gaelic",
  "gl": "Galician",
  "gn": "Guarani",
  "gu": "Gujarati",
  "gv": "Manx",
  "ha": "Hausa",
  "he": "Hebrew",
  "hi": "Hindi",
  "ho": "Hiri Motu",
  "hr": "Croatian",
  "ht": "Haitian |  Haitian Creole",
  "hu": "Hungarian",
  "hy": "Armenian",
  "hz": "Herero",
  "ia": "Interlingua (International Auxiliary Language Association)",
  "id": "Indonesian",
  "ie": "Interlingue |  Occidental",
  "ig": "Igbo",
  "ii": "Sichuan Yi |  Nuosu",
  "ik": "Inupiaq",
  "io": "Ido",
  "is": "Icelandic",
  "it": "Italian",
  "iu": "Inuktitut",
  "ja": "Japanese",
  "jv": "Javanese",
  "ka": "Georgian",
  "kg": "Kongo",
  "ki": "Kikuyu |  Gikuyu",
  "kj": "Kuanyama |  Kwanyama",
  "kk": "Kazakh",
  "kl": "Kalaallisut |  Greenlandic",
  "km": "Central Khmer",
  "kn": "Kannada",
  "ko": "Korean",
  "kr": "Kanuri",
  "ks": "Kashmiri",
  "ku": "Kurdish",
  "kv": "Komi",
  "kw": "Cornish",
  "ky": "Kirghiz |  Kyrgyz",
  "la": "Latin",
  "lb": "Luxembourgish |  Letzeburgesch",
  "lg": "Ganda",
  "li": "Limburgan |  Limburger |  Limburgish",
  "ln": "Lingala",
  "lo": "Lao",
  "lt": "Lithuanian",
  "lu": "Luba-Katanga",
  "lv": "Latvian",
  "mg": "Malagasy",
  "mh": "Marshallese",
  "mi": "Maori",
  "mk": "Macedonian",
  "ml": "Malayalam",
  "mn": "Mongolian",
  "mr": "Marathi",
  "ms": "Malay",
  "mt": "Maltese",
  "my": "Burmese",
  "na": "Nauru",
  "nb": "Bokm\u00e5l, Norwegian |  Norwegian Bokm\u00e5l",
  "nd": "Ndebele, North |  North Ndebele",
  "ne": "Nepali",
  "ng": "Ndonga",
  "nl": "Dutch |  Flemish",
  "nn": "Norwegian Nynorsk |  Nynorsk, Norwegian",
  "no": "Norwegian",
  "nr": "Ndebele, South |  South Ndebele",
  "nv": "Navajo |  Navaho",
  "ny": "Chichewa |  Chewa |  Nyanja",
  "oc": "Occitan (post 1500)",
  "oj": "Ojibwa",
  "om": "Oromo",
  "or": "Oriya",
  "os": "Ossetian |  Ossetic",
  "pa": "Panjabi |  Punjabi",
  "pi": "Pali",
  "pl": "Polish",
  "ps": "Pushto |  Pashto",
  "pt": "Portuguese",
  "qu": "Quechua",
  "rm": "Romansh",
  "rn": "Rundi",
  "ro": "Romanian |  Moldavian |  Moldovan",
  "ru": "Russian",
  "rw": "Kinyarwanda",
  "sa": "Sanskrit",
  "sc": "Sardinian",
  "sd": "Sindhi",
  "se": "Northern Sami",
  "sg": "Sango",
  "si": "Sinhala |  Sinhalese",
  "sk": "Slovak",
  "sl": "Slovenian",
  "sm": "Samoan",
  "sn": "Shona",
  "so": "Somali",
  "sq": "Albanian",
  "sr": "Serbian",
  "ss": "Swati",
  "st": "Sotho, Southern",
  "su": "Sundanese",
  "sv": "Swedish",
  "sw": "Swahili",
  "ta": "Tamil",
  "te": "Telugu",
  "tg": "Tajik",
  "th": "Thai",
  "ti": "Tigrinya",
  "tk": "Turkmen",
  "tl": "Tagalog",
  "tn": "Tswana",
  "to": "Tonga (Tonga Islands)",
  "tr": "Turkish",
  "ts": "Tsonga",
  "tt": "Tatar",
  "tw": "Twi",
  "ty": "Tahitian",
  "ug": "Uighur |  Uyghur",
  "uk": "Ukrainian",
  "ur": "Urdu",
  "uz": "Uzbek",
  "ve": "Venda",
  "vi": "Vietnamese",
  "vo": "Volap\u00fck",
  "wa": "Walloon",
  "wo": "Wolof",
  "xh": "Xhosa",
  "yi": "Yiddish",
  "yo": "Yoruba",
  "za": "Zhuang |  Chuang",
  "zh": "Simplified Chinese",
  "zu": "Zulu",
  "zh-hk": "Traditional Chinese",
  "pt-br": "Brazilian Portugese",
  "es-la": "Latin American Spanish",
  "ja-ro": "Romanized Japanese",
  "ko-ro": "Romanized Korean",
  "zh-ro": "Romanized Chinese"
}
